from elftools.elf.elffile import ELFFile


def valid_segment_to_add(segment):
    return segment['p_type'] == 'PT_LOAD' and segment['p_filesz'] and segment['p_memsz'] and segment['p_memsz'] and (
        2 & segment['p_flags'])
    # return segment['p_type'] == 'PT_LOAD' and segment['p_filesz'] and segment['p_memsz'] and segment['p_memsz']


def get_segments(filename):
    with open(filename, 'rb') as f:
        elffile = ELFFile(f)
        segments = []
        for segment in elffile.iter_segments():
            if valid_segment_to_add(segment):
                segment = {"file_addr_begin": segment["p_offset"],
                           "file_addr_end": segment["p_offset"] + segment["p_filesz"],
                           "virt_addr_begin": segment["p_vaddr"],
                           "virt_addr_end": segment["p_vaddr"] + segment["p_filesz"]}
                segments.append(segment)
    return segments


def get_version(filename):
    with open(filename, 'rb') as f:
        return ELFFile(f).elfclass


def addr_in_range(addr, segment):
    return segment["virt_addr_begin"] <= addr < segment["virt_addr_end"]


def valid_addresses(addresses, segments):
    result = True
    for address in addresses:
        result = result and valid_address(address, segments)
        if not result:
            return False
    return result


def valid_address(addr, segments):
    # first_segment = get_first_segment(segments)
    for segment in segments:
        if addr_in_range(addr, segment):
            # if addr_in_range(addr, segment) and not addr_in_range(addr, first_segment):
            return True
    return False


def valid_pointer(addr, segments):
    for segment in segments:
        if addr_in_range(addr, segment):
            return True


def get_segment_boundaries(addr, segments):
    for segment in segments:
        if addr_in_range(addr, segment):
            return segment


def get_segment_from_offset(offset, segments):
    for segment in segments:
        if segment["file_addr_begin"] <= offset < segment["file_addr_end"]:
            return segment


def get_address_from_offset(offset, segments):
    segment = get_segment_from_offset(offset, segments)
    return offset - segment["file_addr_begin"] + segment["virt_addr_begin"]


def get_offset(address, segments):
    segment = get_segment_boundaries(address, segments)
    return segment["file_addr_begin"] + address - segment["virt_addr_begin"]


# def get_first_segment(segments):
#     for segment in segments:
#         if segment["virt_addr_begin"] == 0:
#             return segment


def get_all_addresses(segments):
    for segment in segments:
        for address in range(segment["virt_addr_begin"], segment["virt_addr_end"], alignment):
            yield address


alignment = 4

if __name__ == '__main__':
    filename = "./dumps/core.22416"
    segments = get_segments(filename)
    # print(addresses)
    print(len(segments))
    version = get_version(filename)
    # print(version)
    # print(valid_address(0xb44534, addresses), valid_address(0xfffffffff600ff0, addresses),
    #       valid_address(0x7fefd054c540, addresses))
    # print(hex(get_offset(0x7fefd054c540, addresses)))
