from struct import unpack, pack

from elf import get_all_addresses, valid_address, get_offset
from php5.entry import get_glob_segments, get_data
from php5.my_types import *
from php5.my_hashtable import ArrayZvalue
from php5.objects.zend_objects_store import ObjectStore
from php5.zend_execute_data import ZendExecuteData

ret_val_pp = 0
uninint_zval_p = ret_val_pp + 8 + zval_size
error_zval_p = uninint_zval_p + 8 + zval_size
sym_tab_c = error_zval_p + 8
sym_tab_c_l = sym_tab_c + 8
sym_tab_c_p = sym_tab_c_l + 8
opline_p = sym_tab_c_p + 8
active_symbol_table = opline_p + 8
symbol_table = active_symbol_table + 8
included_files = symbol_table + hashtable_size
bailout = included_files + hashtable_size
active_op_array = bailout + 24
function_table = active_op_array + 8
class_table = function_table + 8
zend_constants = class_table + 8
regular_list = zend_constants + 72
persistent_list = regular_list + hashtable_size
object_store = persistent_list + hashtable_size + 48 + 112
execute_data = object_store + object_store_size + 16 + 8 + 48 * 3


# execute_data = object_store + object_store_size + 16 + 8 + 48 * 3


class ExecutorGlobals:
    def __init__(self, level=0) -> None:
        self.executor_globals_address = -1
        self.symbol_table, self.active_symbol_table, self.included_files, self.function_table, self.constants_table, \
        self.class_table, self.persistent_list, self.regular_list = (None,) * 8
        # for address in get_all_addresses(get_glob_segments()):
        # for address in [0x7f3fc9a53b30]:
        # for address in [0x7f3fcb88c9c8]:
        # for address in [0x7f3fc9a59ed0]:
        # for address in [0x7f3fcb88c958]:
        # for address in [0x7f5bfa574958]:
        # for address in [0x7f5bfa574958]:
        for address in [0x7fe42d9d3958]:
            if self.valid_excutor_globals_address(address):
                self.executor_globals_address = address
                print(hex(address))
                self.build_executor_globals()
                break
                # else:
                # print(hex(self.executor_globals_address))

    def build_executor_globals(self):
        base = self.executor_globals_address
        self.symbol_table = ArrayZvalue(pack('QQ', symbol_table + base, 0))
        self.included_files = ArrayZvalue(pack('QQ', included_files + base, 0))
        print(self.symbol_table)
        print(self.included_files)
        active_symbol_table_p = self.get_pointer_at_address(active_symbol_table + base)
        self.active_symbol_table = ArrayZvalue(pack('QQ', active_symbol_table_p, 0))
        print(self.active_symbol_table)
        print(self.active_symbol_table.is_empty())

        function_table_p = self.get_pointer_at_address(function_table + base)
        self.function_table = ArrayZvalue(pack('QQ', function_table_p, 0))
        # print(self.function_table)

        class_table_p = self.get_pointer_at_address(class_table + base)
        from php5.objects.zend_class_entry import ZendClassEntry
        self.class_table = ArrayZvalue(pack('QQ', class_table_p, 0), special_type=ZendClassEntry)
        print(self.class_table)

        zend_constants_p = self.get_pointer_at_address(zend_constants + base)
        self.constants_table = ArrayZvalue(pack('QQ', zend_constants_p, 0))
        # print(self.constants_table)

        self.regular_list = ArrayZvalue(pack('QQ', regular_list + base, 0))
        # print(self.regular_list)

        self.persistent_list = ArrayZvalue(pack('QQ', persistent_list + base, 0))
        # print(self.persistent_list)

        self.object_store = ObjectStore(object_store + base)
        print(self.object_store)

        self.execute_data_p = self.get_pointer_at_address(base + execute_data)
        self.execute_data = ZendExecuteData(self.execute_data_p)
        print(self.execute_data)


    def get_objects_store_address(self):
        pass

    def get_pointer_at_address(self, address):
        offset = get_offset(address, get_glob_segments())
        return unpack('Q', get_data(offset, 8))[0]

    def valid_excutor_globals_address(self, address):
        result = True
        result = result and self.check_pointers(address)
        result = result and self.check_hashtables(address)
        return result
        # return self.check_pointers(address) and self.check_hashtables(address)

    def check_pointer(self, address):
        if valid_address(address, get_glob_segments()):
            pointer = self.get_pointer_at_address(address)
            return pointer == 0 or valid_address(pointer, get_glob_segments())
        return False

    def check_pointers(self, address):
        for offset in [ret_val_pp, uninint_zval_p, error_zval_p, sym_tab_c, sym_tab_c_l, sym_tab_c_p,
                       active_symbol_table, bailout, active_op_array, function_table, class_table, zend_constants]:
            if not self.check_pointer(address + offset):
                return False
        return True

    def check_hashtables(self, address):
        for offset in [symbol_table, included_files]:
            if not self.check_hashtable(offset + address):
                return False
        return True

    def check_hashtable(self, address):
        if valid_address(address, get_glob_segments()):
            hashtable_header = ArrayZvalue(pack('QQ', address, 0), header_only=1)
            if not hashtable_header.valid_header() or hashtable_header.is_empty():
                return False
        return True


if __name__ == "__main__":
    # filename = '../dumps/php5/core.14197'
    filename = '../dumps/php5/core.2972'
    filename = '../dumps/php5/core.2552'
    from php5.entry import init

    # 0x7f3fcb88c9c8
    init(filename)
    glob = ExecutorGlobals()
    print("==============================")
    print(glob.object_store.get_bucket_from_object_store_list(1))
    print(glob.object_store)
    # 0x7f3fcb7b4b48

    #
    # String    Virtual    address: 0x7f5bfa49fc48
    # String    Offset: 0x97400c
    # String    Length: 14(0xe)
    # Zval    virtual    address: 0x7f5bfa49ca58
    # Zval    file    Offset: 0x970e1c
    # Refcount: 2
    # IsRef: 0
