from _struct import unpack

from elf import get_offset, valid_address
from php5.entry import get_data, get_glob_segments
from php5.my_types import *


class Bucket:
    def __init__(self, address, level=0, special_type=None):
        self.address, self.level = address, level
        self.special_type = special_type
        self.offset = get_offset(address, get_glob_segments())
        self.data = self.load_bucket_from_file()
        self.parsed = self.parse_data()
        self.value = self.create_value()

    def load_bucket_from_file(self):
        return get_data(self.offset, 72)

    def parse_data(self):
        return unpack('QiI7Q', self.data)

    def get_next(self):
        return self.parsed[pListNext]

    def get_zval(self):
        return self.value

    def create_value(self):
        from php5.my_zval_struct import ZvalStruct
        if self.special_type is None:
            zval2 = ZvalStruct(self.parsed[pDataPtr], self.level + 1)
            zval1 = ZvalStruct(self.parsed[pData], self.level + 1)

        else:
            zval2 = self.special_type(self.parsed[pDataPtr], self.level + 1)
            zval1 = self.special_type(self.parsed[pData], self.level + 1)

        if zval2.is_valid():
            return zval2
        if zval1.is_valid():
            return zval1
        return zval2

    def get_string_key(self):
        from php5.my_zval_struct import get_data_from_file
        if self.parsed[nKeyLength] == 0:
            return None
        elif valid_address(self.parsed[arKey], get_glob_segments()):
            data = get_data(get_offset(self.parsed[arKey], get_glob_segments()), self.parsed[nKeyLength] - 1)
            return data

    def get_data_to_print(self):
        indent = (self.level) * spacer
        string_key = self.parsed[nKeyLength] > 0
        args = {'indent': indent, 'index': self.parsed[hashIndex], 'hindex': hex(self.parsed[hashIndex]),
                'key': self.get_string_key(), 'keyLen': self.parsed[nKeyLength],
                'hkeyLen': hex(self.parsed[nKeyLength]), 'zval': self.get_zval(), 'begin': '>'
            , 'pdata': hex(self.parsed[pData]), 'pdataplus': hex(self.parsed[pData] + 8),
                'pdataptr': hex(self.parsed[pDataPtr]), 'addr': hex(self.address)}
        if values_only:
            if string_key:
                return '{indent}{begin}Key: {key}\n{zval}\n'.format(**args)
                # return '{indent}{begin}Key: {key} {addr} {pdata} {pdataplus} {pdataptr}\n{zval}\n'.format(**args)
            else:
                return '{indent}{begin}Index: {index}({hindex})\n{zval}\n'.format(**args)
        if string_key:
            return '{indent}Type: Bucket\n{indent}Hash: {hindex}\n{indent}KeyLength: {keyLen}({hkeyLen})\n{indent}Key: {key}\n{zval}'.format(
                **args)
        else:
            return '{indent}Type: Bucket\n{indent}Index: {index}({hindex})\n{zval}'.format(**args)

    def __str__(self):
        return self.get_data_to_print()

    def __repr__(self):
        return self.get_data_to_print()


if __name__ == '__main__':
    filename = '../dumps/core.2590'
    from php5.my_zval_struct import ZvalStruct
    from elf import get_segments, get_all_addresses

    segments = get_segments(filename)
    f = open(filename, 'rb')
    address = 0x7f54c8b1e320
    zval = ZvalStruct(f, segments, address, 0)
    print(zval)
    for address in get_all_addresses(segments):
        zval = ZvalStruct(f, segments, address, 0)
        if zval.is_valid() and zval.type == IS_ARRAY:
            print(zval)
        del zval
    f.close()
