from elf import valid_addresses, get_all_addresses, get_segments
from php5.my_bucket import *
from php5.entry import get_glob_segments, get_data


class ArrayZvalue:
    def __init__(self, raw_data, level=0, header_only=0, special_type=None):
        self.raw_data, self.level = raw_data, level
        self.special_type = special_type
        self.header_only = header_only
        self.parsed = (-1,) * 9
        self.valid_parsed = False
        self.valid = False
        self.bucket_list = []

        self.hash_table_address = unpack('Q', self.raw_data[:8])[0]
        if valid_address(self.hash_table_address, get_glob_segments()):
            self.hash_table_offset = get_offset(self.hash_table_address, get_glob_segments())
            self.data = get_data(self.hash_table_offset, hashtable_size)
            self.parsed = self.parse_hashtable()
            self.valid_parsed = self.check_parsed()
            if self.valid_parsed:
                if not self.header_only:
                    self.bucket_list = self.get_list(self.parsed[pListHead])
                self.valid = self.check_hashtable()

    def check_hashtable(self):
        if valid_address(self.hash_table_address, get_glob_segments()) and self.is_empty() and self.valid_parsed:
            return True
        if self.header_only:
            return self.parsed
        return valid_address(self.hash_table_address, get_glob_segments()) and self.bucket_list and self.valid_parsed
        # return valid_address(self.hash_table_address, get_glob_segments()) and self.valid_parsed

    def is_valid(self):
        return self.valid

    def valid_header(self):
        return self.valid_parsed

    def is_empty(self):
        return self.parsed[nTableSize] == self.parsed[nTableMask] == self.parsed[nNumOfElements] == pListHead == \
               self.parsed[pListHead] == self.parsed[pListTail] == 0

    def check_parsed(self):
        return self.parsed[nTableSize] - 1 == self.parsed[nTableMask] and \
               self.parsed[nNumOfElements] <= self.parsed[nTableSize] and \
               valid_addresses(
                   [self.parsed[pListHead], self.parsed[pListTail], self.parsed[arBuckets]],
                   get_glob_segments())

    def get_list(self, first_bucket_address):
        result = []
        actual = Bucket(first_bucket_address, self.level + 1, special_type=self.special_type)
        result.append(actual)
        while True:
            target = actual.get_next()
            if target and valid_address(target, get_glob_segments()):
                target = Bucket(target, self.level + 1, special_type=self.special_type)
                result.append(target)
                actual = target
            else:
                break
        return result

    def get_data_to_print(self):
        indent = self.level * spacer
        if not self.is_valid():
            return indent + 'Invalid Hashtable\n'
        args = {'indent': indent, 'addr': hex(self.hash_table_address),
                'carret': '{', 'ecarret': '}',
                'bucket_string': ''.join([str(x) for x in self.bucket_list])}
        if values_only:
            return '{indent}{carret}\n{bucket_string}{indent}{ecarret}'.format(**args)
        else:
            return '{indent}Type: Hashtable\n{bucket_string}'.format(
                **args)

    def get_bucket_list_size(self):
        return len(self.bucket_list)

    def __repr__(self):
        return self.get_data_to_print()

    def __str__(self):
        return self.get_data_to_print()

    def parse_hashtable(self):
        return unpack(hashtable_format, self.data[:-16])


if __name__ == '__main__':
    # filename = '../dumps/core.2498'
    filename = '../dumps/core.2590'
    filename = '../dumps/php5/core.14197'

    # address = 0x7f54c8b1e320
    from php5.my_zval_struct import ZvalStruct
    from php5.entry import init
    from struct import pack

    init(filename)
    # print(valid_address(address, get_glob_segments()))
    # zval = ZvalStruct(address, 0)
    # print(zval)
    # for address in get_all_addresses(get_glob_segments()):
    #     zval = ZvalStruct(address, 0)
    #     if zval.is_valid() and zval.type == IS_ARRAY:
    #         print(zval)
    #         print('========================================')
    #     del zval
    for address in get_all_addresses(get_glob_segments()):
        zval = ArrayZvalue(pack('QQ', address, 0), 0)
        if zval.is_valid():
            print(zval)
            print('========================================')
        del zval
