from _struct import *
from php5.entry import get_glob_segments, get_data
from elf import *
from php5.my_types import *


class StringZvalue:
    def __init__(self, raw_data, level=0):
        self.raw_data, self.level = raw_data, level
        self.string_address, self.string_offset, self.string_length, self.padding, self.string = self.get_value()
        self.deprecation_size = string_deprecation_size
        self._valid = self.check_string()

    def is_valid(self):
        return self._valid

    def check_string(self):
        return valid_address(self.string_address,
                             get_glob_segments()) and self.deprecation_size > self.string_length > 0

    def get_value(self):
        result = unpack('<Qii', self.raw_data)
        if not valid_address(result[0], get_glob_segments()):
            return -1, -1, -1, -1, ""
        string_offset = get_offset(result[0], get_glob_segments())
        string = get_data(string_offset, result[1])
        return result[0], string_offset, result[1], result[2], string

    def get_string_to_print(self):
        indent = self.level * spacer
        if not self.is_valid():
            return '{}Invalid String\n'.format(indent)
        if values_only:
            return '{indent}{string}'.format(string=self.string, indent=indent)
        args = {'indent': indent, 'string': self.string,
                'sva': hex(self.string_address), 'sfo': hex(self.string_offset), 'sl': self.string_length,
                'hsl': hex(self.string_length)}
        result = '{indent}Type: String\n{indent}Value: {string}\n{indent}String Virtual address: {sva}\n{indent}String Offset: {sfo}\n{indent}String Length: {sl}({hsl})\n'.format(
            **args)
        return result

    def __repr__(self):
        return self.get_string_to_print()

    def __str__(self):
        return self.get_string_to_print()


if __name__ == '__main__':
    filename = '../dumps/core.2498'
    filename = '../dumps/php5/core.14197'
    filename = '../dumps/php5/core.2972'
    filename = '../dumps/php5/core.2552'

    # addr = 0x3113d20
    # addr = 0x7f54c8b1e268
    from php5.my_zval_struct import ZvalStruct
    from php5.entry import init

    init(filename)
    # valid = valid_address(addr, get_glob_segments())
    # print(valid)

    # zval = ZvalStruct(addr, 0)
    # if zval.is_valid():
    #     print(zval)
    for address in get_all_addresses(get_glob_segments()):
        zval = ZvalStruct(address, 0)
        if zval.is_valid() and zval.type == IS_STRING:
            print(zval)
        del zval
