from struct import unpack

from elf import get_offset, get_all_addresses, valid_address
from php5.entry import get_data, get_glob_segments
from php5.my_bool import BoolZvalue
from php5.my_double import DoubleZvalue
from php5.my_hashtable import ArrayZvalue
from php5.my_invalid import InvalidZvalue
from php5.my_long import LongZvalue
from php5.my_resource import ResourceZvalue
from php5.my_string import StringZvalue
from php5.my_types import *
from php5.null_zvalue import NullZvalue
from php5.objects.my_object import ObjectZvalue


def get_data_from_file(f, address, size, segments):
    f.seek(get_offset(address, segments))
    return f.read(size)


class ZvalStruct:
    def __init__(self, address, level=0):
        self.address, self.level = address, level
        self.valid = False
        if valid_address(address, get_glob_segments()):
            self.offset = get_offset(self.address, get_glob_segments())
            self.data = self.load_zval_data()
            self.zvalue, self.valid, self.valid_attributes = None, False, False
            self.zvalue_raw_data, (
                self.refcount__gc, self.type, self.is_ref__gc, self.padding) = self.parse_zval_struct()
            self.valid_attributes = self.check_attributes()
            if self.valid_attributes:
                self.zvalue = self.create_zvalue()
            self.valid = self.valid_attributes and self.zvalue is not None and self.zvalue.is_valid()

    def check_attributes(self):
        return self.refcount__gc > 0 and 0 <= self.type < 10 and (
            self.is_ref__gc in [0, 1]) and self.refcount__gc < deprecation_refcount
        # return deprecation_refcount > self.refcount__gc > 0 and self.__type == IS_DOUBLE and (self.is_ref__gc in [0, 1])
        # return deprecation_refcount > self.refcount__gc > 0 and self.__type == IS_STRING and (self.is_ref__gc in [0, 1])
        # return deprecation_refcount > self.refcount__gc > 0 and self.__type in [IS_CONSTANT,IS_CONSTANT_AST] and (self.is_ref__gc in [0, 1]) and self.padding == 0
        # return deprecation_refcount > self.refcount__gc > 0 and self.__type == IS_ARRAY and (self.is_ref__gc in [0, 1])

    def create_zvalue(self):
        result = None
        if self.type == IS_NULL:
            result = NullZvalue(self.zvalue_raw_data, self.level)
        elif self.type == IS_LONG:
            result = LongZvalue(self.zvalue_raw_data, self.level)
        elif self.type == IS_DOUBLE:
            result = DoubleZvalue(self.zvalue_raw_data, self.level)
        elif self.type == IS_BOOL:
            result = BoolZvalue(self.zvalue_raw_data, self.level)
        elif self.type == IS_ARRAY:
            result = ArrayZvalue(self.zvalue_raw_data, self.level)
        elif self.type == IS_OBJECT:
            result = ObjectZvalue(self.zvalue_raw_data, self.level)
        elif self.type == IS_STRING:
            result = StringZvalue(self.zvalue_raw_data, self.level)
        elif self.type == IS_RESOURCE:
            result = ResourceZvalue(self.zvalue_raw_data, self.level)
        elif self.type == IS_CONSTANT:
            pass
            # result = ConstantZvalue(self, self.level)
        elif self.type == IS_CONSTANT_AST:
            # print(hex(self.address))
            pass
        else:
            result = InvalidZvalue()
        # if result is None:
        #     return NullZvalue(self, self.level)
        return result

    def is_valid(self):
        return self.valid

    def __repr__(self):
        return self.get_data_to_print()

    def __str__(self):
        return self.get_data_to_print()

    def get_info(self):
        indent = self.level * spacer
        args = {'indent': indent, 'isref': self.is_ref__gc, 'refcount': self.refcount__gc, 'virt': hex(self.address),
                'offset': hex(self.offset)}
        return '{indent}Zval virtual address: {virt}\n{indent}Zval file Offset: {offset}\n{indent}Refcount: {refcount}\n{indent}IsRef: {isref}\n'.format(
            **args)

    def load_zval_data(self):
        return get_data(get_offset(self.address, get_glob_segments()), 24)
        # return get_data_from_file(self.f, self.address, 24, self.segments)

    def parse_zval_struct(self):
        return self.data[:16], unpack('IBBh', self.data[16:24])

    def get_data_to_print(self):
        indent = self.level * spacer
        if not self.valid:
            return indent + 'Invalid zval\n'
        if values_only:
            return self.zvalue.__str__()
        return self.zvalue.__str__() + self.get_info()


if __name__ == '__main__':
    filename = 'dumps/core.2498'
    # filename = '../dumps/core.2590'
    filename = '../dumps/php5/core.2590'
    num = 0
    from php5.entry import init, get_glob_segments

    init(filename)
    # address = 0x7f54c8b1e350
    # zval = ZvalStruct(f,segments, address, 0)
    # if zval.is_valid():
    #     print(zval)

    for address in get_all_addresses(get_glob_segments()):
        zval = ZvalStruct(address, 0)
        if zval.is_valid():
            print(zval)
            num += 1
        del zval
    # for address in get_all_addresses(segments):
    #     zval = ZvalStruct(f, segments, address, 0)
    #     if zval.is_valid() and zval.__type == IS_ARRAY:
    #         print(zval)
    #         num += 1
    #     del zval
    print(num)
