from struct import unpack
from elf import valid_address, get_all_addresses
from php5.entry import get_glob_segments
from php5.my_types import IS_OBJECT, spacer


class ObjectZvalue:
    def __init__(self, raw_data, level) -> None:
        self.raw_data, self.level = raw_data, level
        self.valid = False
        self.index, padding, self.object_handlers = self.parse_object()
        if self.valid_parsed():
            # TODO get from object bucket store
            self.valid = True

    def is_valid(self):
        return self.valid

    def parse_object(self):
        return unpack('iiQ', self.raw_data)

    def valid_parsed(self):
        return valid_address(self.object_handlers, get_glob_segments())

    def get_data_to_print(self):
        indent = self.level * spacer
        if not self.valid:
            return '{}Invalid object\n'.format(indent)
        return '{}Object index: {}\n'.format(indent, hex(self.index))

    def __repr__(self) -> str:
        return self.get_data_to_print()

    def __str__(self) -> str:
        return self.get_data_to_print()


if __name__ == '__main__':
    filename = '../dumps/core.2498'
    from php5.my_zval_struct import ZvalStruct
    from php5.entry import init

    init(filename)
    # for address in get_all_addresses(get_glob_segments()):
    for address in [0x7f551e08a958]:
        zval = ZvalStruct(address, 0)
        if zval.is_valid() and zval.type == IS_OBJECT:
            print(zval, hex(address))
        del zval
