from elf import valid_address, get_offset, valid_addresses
from php5.entry import get_glob_segments, get_data
from struct import unpack
from php5.my_hashtable import ArrayZvalue
from php5.my_types import zval_size, spacer
from php5.objects.zend_class_entry import ZendClassEntry


class ZendObject:
    def __init__(self, address, level=0) -> None:
        self.address, self.level = address, level
        self.valid = False
        if valid_address(self.address, get_glob_segments()):
            self.offset = get_offset(self.address, get_glob_segments())
            self.data = get_data(self.offset, 24)
            self.class_entry_pointer, self.properties_pointer, self.properties_table_pointer = self.parse_object()
            if valid_addresses([self.class_entry_pointer, self.properties_pointer, self.properties_table_pointer],
                               get_glob_segments()):
                self.class_entry = ZendClassEntry(self.class_entry_pointer, self.level + 1)
                self.properties = ArrayZvalue(self.data[8:16], self.level + 1)
                self.properties_table = []
                if self.class_entry.is_valid():
                    self.properties_num = self.class_entry.get_properties_info_size()
                    self.properties_table = self.get_zval_array(self.properties_table_pointer, self.properties_num)
                    # print(self)
                self.valid = self.check_valid()

    def is_valid(self):
        return self.valid

    def get_data_to_print(self):
        indent1 = self.level * spacer
        indent2 = (self.level + 1) * spacer
        if not self.valid:
            return '{}Invalid object\n'.format(indent1)
        args = {'ce': self.class_entry, 'prop': self.properties, 'indent1': indent1, 'indent2': indent2,
                'pt': '\n'.join([str(x) for x in self.properties_table])}
        return '{indent1}Type object:\n{indent2}Class entry:\n{ce}{indent2}Properties:\n{prop}\n{indent2}Properties table:\n{pt}\n'.format(
            **args)

    def __repr__(self) -> str:
        return self.get_data_to_print()

    def __str__(self) -> str:
        return self.get_data_to_print()

    def parse_object(self):
        return unpack('QQQ', self.data)

    def check_valid(self):
        return self.class_entry.is_valid() and self.properties.is_valid()

    def get_pointer_at_address(self, address):
        offset = get_offset(address, get_glob_segments())
        return unpack('Q', get_data(offset, 8))[0]

    def get_zval_array(self, address, count):
        result = []
        for index in range(count):
            pointer = self.get_pointer_at_address(address + 8 * index)
            # pointer = address + 8 * index
            from php5.my_zval_struct import ZvalStruct
            if valid_address(pointer, get_glob_segments()):
                addr2 = self.get_pointer_at_address(pointer)
            else:
                addr2 = 0
            zval = ZvalStruct(addr2, self.level + 2)
            result.append(zval)
        return result
