from struct import unpack

from elf import valid_address, get_offset
from php5.entry import get_glob_segments, get_data
from php5.my_types import spacer
from php5.objects.zend_object import ZendObject


class ObjectStoreBucket:
    def __init__(self, address, level) -> None:
        self.address, self.level = address, level
        self.valid = False
        if valid_address(self.address, get_glob_segments()):
            self.offset = get_offset(self.address, get_glob_segments())
            self.data = get_data(self.offset, 16)
            (
                self.destructor_called, self.bucket_used,
                self.apply_count), self.object_pointer = self.parse_store_bucket()
            if self.bucket_used == 1:
                if valid_address(self.object_pointer, get_glob_segments()):
                    self.object = ZendObject(self.object_pointer, self.level + 1)
                    self.valid = self.check_valid()
                    print(self.object)
            elif self.bucket_used == 0:
                self.valid = self.check_valid()
            else:
                pass

    def is_valid(self):
        return self.valid

    def check_valid(self):
        if self.destructor_called not in [0, 1] or self.bucket_used not in [0, 1]:
            return False
        if self.bucket_used:
            return self.object.is_valid()
        else:
            return False

    def is_object_present(self):
        return self.bucket_used

    def parse_store_bucket(self):
        return unpack('bbB', self.data[:3]), unpack('Q', self.data[8:])[0]

    def get_object(self):
        return self.object

    def get_data_to_print(self):
        indent = self.level * spacer
        if not self.valid:
            return '{}Invalid Object Store Bucket\n'.format(indent)
        if self.bucket_used == 0:
            return '{}Next Object Store Bucket\n'.format(indent)
        return self.get_object().__str__()

    def __repr__(self) -> str:
        return self.get_data_to_print()

    def __str__(self) -> str:
        return self.get_data_to_print()
