# typedef struct _zend_objects_store {
# 	zend_object_store_bucket *object_buckets;
# 	zend_uint top;
# 	zend_uint size;
# 	int free_list_head;
# } zend_objects_store;
from struct import unpack

from elf import valid_address, get_offset
from php5.entry import get_glob_segments, get_data
from php5.my_types import object_store_size, object_store_format, object_store_bucket_size, spacer
from php5.objects.zend_object_store_bucket import ObjectStoreBucket


class ObjectStore:
    def __init__(self, address, level=0) -> None:
        self.address, self.level = address, level
        self.valid = False
        self.object_store_bucket_list = []
        if valid_address(self.address, get_glob_segments()):
            self.offset = get_offset(self.address, get_glob_segments())
            self.data = get_data(self.offset, object_store_size)
            self.object_store_bucket_pointer, self.top, self.size, self.free_list_head, pad = self.parse_object_store()
            if valid_address(self.object_store_bucket_pointer, get_glob_segments()):
                self.object_store_bucket_list = self.create_object_store_bucket_list()
                self.valid = True

    def is_valid(self):
        return self.valid

    def parse_object_store(self):
        return unpack(object_store_format, self.data)

    def create_object_store_bucket_list(self):
        result = []
        for index in range(self.size):
            address = self.object_store_bucket_pointer + index * object_store_bucket_size
            object = ObjectStoreBucket(address, self.level + 2)
            if object.is_valid():
                result.append((index, object.get_object()))
        return result

    def get_bucket_from_object_store_list(self, index):
        for index2, value in self.object_store_bucket_list:
            if index == index2:
                return value
        return None

    def get_data_to_print(self):
        indent1 = self.level * spacer
        indent2 = (self.level + 1) * spacer
        if not self.valid:
            return '{}Invalid object store\n'.format(indent1)
        store = '\n'.join([str(indent2) + "Index: " + str(hex(index)) + '\n' + str(value) for index, value in
                           self.object_store_bucket_list])
        args = {'indent1': indent1, 'store': store}
        return '{indent1}Object store:\n{store}\n'.format(**args)

    def __str__(self) -> str:
        return self.get_data_to_print()

    def __repr__(self) -> str:
        return self.get_data_to_print()
