from struct import unpack
from elf import valid_address, get_offset
from php5.entry import get_glob_segments, get_data
from php5.my_types import *


class ZendPropertyInfo:
    def __init__(self, raw_data, level) -> None:
        self.level = level
        self.raw_data = raw_data
        self.address = unpack('Q', raw_data)
        self.valid_address = valid_address(self.address, get_glob_segments())
        self.valid = False
        self.entry = None
        self.name = ''
        self.offset = -1
        if self.valid_address:
            self.property_offset = get_offset(self.address, get_glob_segments())
            self.data = get_data(self.property_offset, property_info_size)
            self.flags, pad1, self.name_p, self.name_length, pad2, self.hash, self.offset, pad3, self.comment_p, self.comment_len, pad4, self.entry_p = self.parse_property_info()
            if self.valid_pointer_len(self.name_p, self.name_length):
                self.name = self.get_string(self.name_p, self.name_length)
            self.valid = self.check_valid()

    def parse_property_info(self):
        return unpack(property_format, self.data)

    def valid_pointer_len(self, pointer, length):
        return valid_address(pointer, get_glob_segments()) and 0 <= length < string_deprecation_size

    def null_pointer_len(self, pointer, length):
        return pointer == length == 0

    def get_string(self, pointer, length):
        return get_data(get_offset(pointer, get_glob_segments()), length)

    def check_valid(self):
        result = True
        result = result and (
            self.valid_pointer_len(self.name_p, self.name_length) or self.null_pointer_len(self.name_p,
                                                                                           self.name_length))
        result = result and (self.entry_p == 0 or valid_address(self.entry_p, get_glob_segments()))
        return result

    def is_valid(self):
        return self.valid

    def get_name(self):
        return self.name

    def get_name_len(self):
        return self.name_length

    def get_data_to_print(self):
        indent = self.level * spacer
        if not self.is_valid():
            return '{}Invlaid property\n'.format(indent)
        else:
            return '{}Name: {} Offset: {}\n'.format(indent, self.name, self.offset)
