from struct import unpack

from elf import valid_address, get_offset
from php5.entry import get_glob_segments, get_data
from php5.my_types import op_array_size_partial, compiled_size, execute_data_size, spacer
from php5.my_zval_struct import ZvalStruct
from php5.sym_tab.compiled_var import CompiledVar


class ZendOpArray:
    def __init__(self, address, ex_data_a, ht_unused, level=0) -> None:
        self.level = level
        self.ex_data_a = ex_data_a
        self.address = address
        self.valid = False
        self.ht_unused = ht_unused
        self.fname = ''
        if valid_address(self.address, get_glob_segments()):
            self.offset = get_offset(self.address, get_glob_segments())
            self.data = get_data(self.offset, op_array_size_partial)
            (self.fname_p, self.scope, pad, self.proto), (self.compiled_p, self.compiled_n) = self.parse_oparray()
            valid_parsed = self.valid_parsed()
            if valid_parsed:
                self.fname = self.get_fname()
                if self.ht_unused:
                    if (self.compiled_p == self.compiled_n == 0 or valid_address(self.compiled_p, get_glob_segments())):
                        self.compiled = self.get_compiled()
                        self.values = self.get_values()
                self.valid = True

    def is_valid(self):
        return self.valid

    def parse_oparray(self):
        return unpack('QQQQ', self.data[8:40]), unpack('Qi', self.data[-12:])

    def valid_parsed(self):
        result = True
        result = result and self.valid_ponter(self.fname_p)
        result = result and self.valid_ponter(self.scope)
        result = result and self.valid_ponter(self.proto)
        return result

    def valid_ponter(self, address):
        return address == 0 or valid_address(address, get_glob_segments())

    def get_fname(self):
        if self.fname_p:
            func_offset = get_offset(self.fname_p, get_glob_segments())
            length = 0
            while unpack('B', get_data(func_offset + length, 1))[0] is not 0:
                length += 1
            return get_data(func_offset, length)
        else:
            return ''

    def get_compiled(self):
        result = []
        for index in range(self.compiled_n):
            x = CompiledVar(self.compiled_p + index * compiled_size, self.level)
            if x.is_valid():
                result.append(x)
        return result

    def get_pointer_at_address(self, address):
        offset = get_offset(address, get_glob_segments())
        return unpack('Q', get_data(offset, 8))[0]

    def get_values(self):
        result = []
        pointer = (self.ex_data_a + execute_data_size + 8)
        # print(hex(pointer))
        pointer2 = self.get_pointer_at_address(pointer)
        for i in range(self.compiled_n):
            pointer = self.get_pointer_at_address(pointer2 + i * 8)
            # print(hex(pointer))
            # pointer3 = self.get_pointer_at_address(pointer)
            # print(hex(pointer))
            result.append(ZvalStruct(pointer, self.level + 1))
        return result

    def get_data_to_print(self):
        indent = self.level * spacer
        indent2 = (self.level + 1) * spacer
        if not self.valid:
            return '{}Invalid op arrray\n'.format(indent)
        merged = '\n'.join([str(indent) + str(name) + '\n' + str(value) for name, value in
                            zip(self.compiled, self.values)]) if self.ht_unused else ''
        args = {'ind': indent, 'ind2': indent2, 'fn': self.fname, 'merged': merged}
        return '{ind}Function name: {fn}\n{merged}'.format(**args)

    def __repr__(self) -> str:
        return self.get_data_to_print()

    def __str__(self) -> str:
        return self.get_data_to_print()
