from struct import unpack, pack

from elf import valid_address, get_offset
from php5.entry import get_glob_segments, get_data
from php5.my_hashtable import ArrayZvalue
from php5.my_types import execute_data_size, spacer
from php5.objects.zend_class_entry import ZendClassEntry
from php5.sym_tab.zend_op_array import ZendOpArray


class ZendExecuteData:
    def __init__(self, address, level=0) -> None:
        self.level = level
        self.address = address
        self.valid = False
        if valid_address(self.address, get_glob_segments()):
            self.offset = get_offset(self.address, get_glob_segments())
            self.data = get_data(self.offset, execute_data_size)
            (self.opline, self.func_state1, self.func_state2, self.op_array_p, self.object, self.symbol_table_p,
             self.prev_exec_data_p, self.old_error, self.nested), (
                self.orig_ret_val, self.cur_scope_p, self.called_scope_p, self.this, self.fast_ret, self.delay_ex,
                self.call_slots, self.call) = self.parse_execute_data()
            self.valid_pointers = self.valid_addresses()
            if self.valid_pointers and self.nested in [0, 1]:
                partial_valid = True

                if self.symbol_table_p:
                    self.symbol_table = ArrayZvalue(pack('QQ', self.symbol_table_p, 0), self.level + 1)
                    partial_valid = partial_valid and (self.symbol_table.is_valid() or self.symbol_table.is_empty())

                if self.op_array_p:
                    self.op_array = ZendOpArray(self.op_array_p, self.address, self.symbol_table_p == 0, self.level)
                # if self.cur_scope_p:
                #     self.cur_scope = ZendClassEntry(self.cur_scope_p, self.level + 1)
                #     partial_valid = partial_valid and (self.cur_scope.is_valid())
                #
                # if self.called_scope_p:
                #     self.called_scope_ = ZendClassEntry(self.called_scope_p, self.level + 1)
                #     partial_valid = partial_valid and (self.called_scope_p.is_valid())

                if self.prev_exec_data_p:
                    self.prev_exec_data = ZendExecuteData(self.prev_exec_data_p, self.level)
                    partial_valid = partial_valid and self.prev_exec_data.is_valid()

                self.valid = partial_valid

    def parse_execute_data(self):
        return unpack('8Qb', self.data[:65]), unpack('8Q', self.data[-64:])

    def valid_pointer(self, address):
        return address == 0 or valid_address(address, get_glob_segments())

    def is_valid(self):
        return self.valid

    def valid_addresses(self):
        result = True
        result = result and self.valid_pointer(self.opline)
        result = result and self.valid_pointer(self.func_state1)
        result = result and self.valid_pointer(self.func_state2)
        result = result and self.valid_pointer(self.op_array_p)
        result = result and self.valid_pointer(self.object)
        result = result and self.valid_pointer(self.symbol_table_p)
        result = result and self.valid_pointer(self.prev_exec_data_p)
        result = result and self.valid_pointer(self.old_error)

        # result = result and self.valid_pointer(self.orig_ret_val)
        # result = result and self.valid_pointer(self.cur_scope_p)
        result = result and self.valid_pointer(self.called_scope_p)
        result = result and self.valid_pointer(self.this)
        # result = result and self.valid_pointer(self.fast_ret)
        result = result and self.valid_pointer(self.delay_ex)
        result = result and self.valid_pointer(self.call_slots)
        result = result and self.valid_pointer(self.call)
        return result

    def get_data_to_print(self):
        indent = self.level * spacer
        if not self.is_valid():
            return '{}Invalid execute data\n'.format(indent)
        prev = '' if self.prev_exec_data_p == 0 else self.prev_exec_data
        st = '' if self.symbol_table_p == 0 else self.symbol_table
        args = {'indent': indent, 'st': st, 'prev': prev, 'oa': self.op_array}
        return '{indent}Execute Data\n{oa}{st}{prev}'.format(**args)

    def __repr__(self) -> str:
        return self.get_data_to_print()

    def __str__(self) -> str:
        return self.get_data_to_print()
