from struct import pack

from elf import valid_address, get_offset
from php7.at_start.entry import get_glob_segments, get_data
from php7.at_start.my_globals import *


class ConstantInTable:
    def __init__(self, zvalue_raw_data, level=0) -> None:
        self.__zvalue_raw_data = zvalue_raw_data
        self.__level = level
        self.__address = unpack(zvalue_format_p, self.__zvalue_raw_data)[0]
        self.__valid_address = valid_address(self.__address, get_glob_segments())
        self.__valid = False
        if self.__valid_address:
            self.__zval_address = self.__address
            from php7.zval_struct import ZvalStruct
            from php7.string_zvalue import StringZvalue
            self.__zval = ZvalStruct(self.__address, self.__level + 1)
            self.__zval_offset = get_offset(self.__zval_address, get_glob_segments())
            self.__zend_string_offset = self.__zval_offset + zval_size
            self.__zend_string = StringZvalue(get_data(self.__zend_string_offset, 8), self.__level + 1)
            self.__valid = self.__check_valid()

    def is_valid(self):
        return self.__valid

    def __check_valid(self):
        return self.__zval.is_valid() and self.__zend_string.is_valid() and self.__zval.get_type() in [IS_STRING,
                                                                                                       IS_LONG, IS_TRUE,
                                                                                                       IS_FALSE,
                                                                                                       IS_DOUBLE]

    def __get_data_to_print(self):
        indent = self.__level * spacer
        if not self.is_valid():
            return '{}Invalid Constant\n'.format(indent)
        string = None if self.__zend_string is None else self.__zend_string.get_string()
        args = {'indent': indent, 'h_address': hex(self.__address), 'h_offset': hex(self.__zval_offset),
                'address': int(self.__address), 'offset': int(self.__zval_offset), 'key': string, 'value': self.__zval,
                'zs': self.__zend_string}
        if values_only:
            return '{indent}>Name: {key}\n{value}'.format(**args)
            # return '{indent}>Name: {key}\n{value}{h_address}'.format(**args)
        else:
            pass
            return '{indent}Type: Constant\n{indent}Name: \n{zs}{indent}Value:\n{value}'.format(**args)

    def __str__(self) -> str:
        return self.__get_data_to_print()

    def __repr__(self) -> str:
        return self.__get_data_to_print()


if __name__ == "__main__":
    filename = '../dumps/php7/core.7058'
    filename = '../dumps/php7/core.3175'
    filename = '../dumps/php7/core.6128'
    from php7.at_start.entry import init

    init(filename)
    from php7.zval_struct import ZvalStruct

    # for address in [0x563f7bb0e740, 0x7fa93a2629e0, 0x7fa93a269490, 0x7fa93a27cae8]:
    # for address in get_all_addresses(get_glob_segments()):
    #     zval = ConstantInTable(pack('Q', address), 0)
    #     if zval.is_valid():
    #         print(zval)

    addr = 0x563f7babd040
    addr = 0x5600ff3c1df0

    zval = ConstantInTable(pack('Q', addr), 0)
    print(zval)

    # for address in get_all_addresses(get_glob_segments()):
    #     zval = ZvalStruct(address, 0)
    #     if zval.is_valid() and zval.get_type() == IS_CONSTANT:
    #         print(zval)
