from struct import unpack

from elf import valid_address
from php7.at_start.entry import get_glob_segments
from php7.at_start.my_globals import zvalue_format_p, spacer
from php7.string_zvalue import StringZvalue


class ArgInfo:
    def __init__(self, raw_data, level) -> None:
        self.__zvalue_raw_data = raw_data
        self.__level = level
        self.__address = unpack(zvalue_format_p, self.__zvalue_raw_data)[0]
        self.__valid_address = valid_address(self.__address, get_glob_segments())
        self.__valid = False
        if self.__valid_address:
            self.__name = StringZvalue(self.__zvalue_raw_data, self.__level + 1)
            self.__valid = self.__check_valid()

    def __check_valid(self):
        return self.__name.is_valid()

    def is_valid(self):
        return self.__valid

    def get_name(self):
        return self.__name.get_string()

    def __get_data_to_print(self):
        indent = self.__level * spacer
        if not self.__valid:
            return '{}Invalid arg info\n'.format(indent)
        # if values_only:
        return '{}Argument name:\n{}'.format(indent, self.__name.get_string())
        # else:
        #     return ''

    def __repr__(self) -> str:
        return self.__get_data_to_print()

    def __str__(self) -> str:
        return self.__get_data_to_print()
