from struct import pack

from elf import valid_address, get_offset
from php7.at_start.entry import get_glob_segments, get_data
from php7.at_start.my_globals import *
from php7.functions.arg_info import ArgInfo
from php7.string_zvalue import StringZvalue


class FunctionZvalue:
    def __init__(self, raw_data, level=0) -> None:
        self.__zvalue_raw_data = raw_data
        self.__level = level
        self.__address = unpack(zvalue_format_p, self.__zvalue_raw_data)[0]
        self.__valid_address = valid_address(self.__address, get_glob_segments())
        self.__valid = False
        self.__arg_info = []
        if self.__valid_address:
            self.__offset = get_offset(self.__address, get_glob_segments())
            self.__data = get_data(self.__offset, function_size_partial)
            # self.__function_type, self.__quick_flags, (self.__function_type2, self.__arg_flags1, self.__arg_flags2, self.__arg_flags3, self.__fn_flags, self.__function_name_p, self.__scope, self.__prototype, self.__num_args, self.__required_num_args, self.__arg_info_p) = self.__parse_function()
            (self.__function_type2, self.__arg_flags1, self.__arg_flags2, self.__arg_flags3, self.__fn_flags,
             self.__function_name_p, self.__scope, self.__prototype, self.__num_args, self.__required_num_args,
             self.__arg_info_p) = self.__parse_function()
            self.__function_name = StringZvalue(pack('Q', self.__function_name_p), self.__level + 1)
            self.__valid = self.__check_valid()
            # if self.is_valid():
            #     self.__arg_info = self.__get_arg_info()

    def is_valid(self):
        return self.__valid

    def __check_valid(self):
        result = True
        result = result and (self.__function_name_p == 0 or self.__function_name.is_valid())
        # result = result and (self.__function_name.is_valid())

        # result = result and (
        #     (self.__arg_info_p == self.__num_args == 0) or valid_address(self.__arg_info_p, get_glob_segments()))
        result = result and (self.__required_num_args <= self.__num_args)
        result = result and (self.__prototype == 0 or valid_address(self.__prototype, get_glob_segments()))
        result = result and (self.__scope == 0 or valid_address(self.__scope, get_glob_segments()))
        # result = result and (valid_address(self.__scope, get_glob_segments()))
        # if result:
        #     from php7.class_entry_zvalue import ZendClassEntry
        #     result = ZendClassEntry(pack('Q',self.__scope)).is_valid()
        # print(hex(self.__arg_info_p), self.__num_args)
        return result

    def __get_data_to_print(self):
        indent = self.__level * spacer
        args = {}
        if not self.is_valid():
            return '{}InvalidFunction\n'.format(indent)
        return '{}Function: {}\n{}\n'.format(indent, self.__function_name.get_string(),
                                             '\n'.join([str(x) for x in self.__arg_info]))
        # if values_only:
        #     pass
        # else:
        #     return ''
        # return ''

    def __str__(self) -> str:
        return self.__get_data_to_print()

    def __repr__(self) -> str:
        return self.__get_data_to_print()

    def __parse_function(self):
        type1 = unpack('B', self.__data[:1])[0]
        qf = unpack('I', self.__data[4:8])[0]
        # return type1, qf, unpack('4BI3Q2IQ', self.__data[8:])
        return unpack('4BI3Q2IQ', self.__data[:-8])

    def __get_arg_info(self):
        result = []
        for index in range(self.__num_args):
            arg_info = ArgInfo(pack('Q', self.__arg_info_p + index * arg_info_size), self.__level + 1)
            if arg_info.is_valid():
                result.append(arg_info)
        return result


if __name__ == "__main__":
    filename = '../../dumps/php7/core.6128'
    filename = '../../dumps/php7/core.28039'
    from php7.at_start.entry import init
    from elf import get_all_addresses

    init(filename)

    # address = 0x5600ff3bfac8
    # zval = FunctionZvalue(pack('Q', address))
    # print(zval)
    # address = 0x5600fe457048
    # zval = StringZvalue(pack('Q', address))
    # print(zval)
    # address = 94562283279632
    # from php7.zval_struct import ZvalStruct
    # zval = ZvalStruct(address)
    # print(zval)
    # address = 94562282072592
    # zval = FunctionZvalue(pack('Q', address))
    # print(zval)
    for address in get_all_addresses(get_glob_segments()):
        zval = FunctionZvalue(pack('Q', address))
        if zval.is_valid():
            print(zval)
        del zval
