from struct import pack

from elf import valid_address, get_offset
from php7.at_start.entry import get_glob_segments, get_data
from php7.at_start.my_globals import *
from php7.functions.arg_info import ArgInfo
from php7.string_zvalue import StringZvalue


class ZendOpArray:
    def __init__(self, raw_data, level=0) -> None:
        self.__zvalue_raw_data = raw_data
        self.__level = level
        self.__address = unpack(zvalue_format_p, self.__zvalue_raw_data)[0]
        self.__valid_address = valid_address(self.__address, get_glob_segments())
        self.__valid = False
        self.__arg_info = []
        if self.__valid_address:
            self.__offset = get_offset(self.__address, get_glob_segments())
            self.__data = get_data(self.__offset, oparray_partial_size)
            (self.__number, pad, self.__vars_pp), (
            self.__function_name_p, self.__scope, self.__prototype, self.__num_args, self.__required_num_args,
            self.__arg_info_p) = self.__parse_function()
            self.__function_name = StringZvalue(pack('Q', self.__function_name_p), self.__level + 1)
            self.__valid = self.__check_valid()
            if self.is_valid():
                self.__arg_info = self.__get_arg_info()
                self.st_names = self.get_names()

    def is_valid(self):
        return self.__valid

    def __check_valid(self):
        result = True
        result = result and (self.__function_name.is_valid())
        result = result and (
            (self.__arg_info_p == self.__num_args == 0) or valid_address(self.__arg_info_p, get_glob_segments()))
        result = result and (self.__required_num_args <= self.__num_args)
        result = result and (self.__prototype == 0 or valid_address(self.__prototype, get_glob_segments()))
        result = result and (self.__scope == 0 or valid_address(self.__scope, get_glob_segments()))
        return result

    def __get_data_to_print(self):
        indent = self.__level * spacer
        args = {}
        if not self.is_valid():
            return '{}InvalidFunction(OA)\n'.format(indent)
        return '{}Function(OA): {}\n{}\n'.format(indent, self.__function_name.get_string(),
                                                 '\n'.join([str(x) for x in self.__arg_info]))

    def __str__(self) -> str:
        return self.__get_data_to_print()

    def __repr__(self) -> str:
        return self.__get_data_to_print()

    def __parse_function(self):
        return unpack('iIQ', self.__data[-16:]), unpack('3Q2IQ', self.__data[8:48])

    def __get_arg_info(self):
        result = []
        for index in range(self.__num_args):
            arg_info = ArgInfo(pack('Q', self.__arg_info_p + index * arg_info_size), self.__level + 1)
            if arg_info.is_valid():
                result.append(arg_info)
        return result

    def get_pointer_at_address(self, address):
        offset = get_offset(address, get_glob_segments())
        return unpack('Q', get_data(offset, 8))[0]

    def get_names(self):
        result = []
        if not valid_address(self.__vars_pp, get_glob_segments()):
            return result
        for index in range(self.__number):
            pointer = self.get_pointer_at_address(self.__vars_pp + 8 * index)
            x = StringZvalue(pack('Q', pointer), self.__level + 1)
            if x.is_valid():
                result.append(x)
        return result


if __name__ == "__main__":
    filename = '../dumps/php7/core.6128'
    from php7.at_start.entry import init
    from elf import get_all_addresses

    init(filename)

    for address in get_all_addresses(get_glob_segments()):
        zval = ZendOpArray(pack('Q', address))
        if zval.is_valid():
            print(zval)
        del zval
