from elf import get_offset, valid_address
from php7.at_start.entry import get_data
from php7.at_start.my_globals import *
from php7.string_zvalue import StringZvalue
from php7.at_start.entry import get_glob_segments


class Bucket:
    def __init__(self, address, level=0) -> None:
        self.__address = address
        self.__level = level
        self.__zend_string_key = None
        self.__string_key_used = False
        self.__valid = False
        if valid_address(self.__address, get_glob_segments()):
            self.__offset = get_offset(self.__address, get_glob_segments())
            self.__data = get_data(self.__offset, bucket_size)
            self.__zval, self.__h, self.__raw_key_pointer, self.__key_pointer = self.__parse_bucket()

            # if self.__zval.is_valid():
            if self.__zval.is_valid() and not self.__zval.get_type() == IS_UNDEF:
                if self.__key_pointer:
                    self.__zend_string_key = StringZvalue(self.__raw_key_pointer, self.__level + 1)
                    if self.__zend_string_key.is_valid():
                        self.__valid = True
                        self.__string_key_used = True
                else:
                    self.__valid = True

    def __parse_bucket(self):
        from php7.zval_struct import ZvalStruct
        zval = ZvalStruct(self.__address, self.__level + 1)
        hash_index, pointer = unpack('QQ', self.__data[zval_size:])
        return zval, hash_index, self.__data[-8:], pointer

    def is_valid(self):
        return self.__valid

    def __check_bucket(self):
        if self.__zval.is_valid():
            if self.__check_undefined() or (self.__zend_string_key is not None and self.__zend_string_key.is_valid()):
                return True
        else:
            return False

    def __check_undefined(self):
        return self.__zval.get_type() == IS_UNDEF

    def __get_data_to_print(self):
        indent = self.__level * spacer
        result = ''
        string = None if self.__zend_string_key is None else self.__zend_string_key.get_string()
        args = {'indent': indent, 'begin': '>', 'zval': self.__zval, 'index_hash': self.__h,
                'h_index_hash': hex(self.__h), 'string': string, 'zend_string': self.__zend_string_key,
                'vadr': hex(self.__address),
                'off': hex(self.__offset)}
        if not self.is_valid():
            return '{indent}Invalid Bucket'.format(**args)
        if values_only:
            if self.__string_key_used:
                result = '{indent}{begin}Key: {string}\n{zval}\n'.format(**args)
            else:
                result = '{indent}{begin}Index: {index_hash}({h_index_hash})\n{zval}\n'.format(**args)
        else:
            if self.__string_key_used:
                result = '{indent}Type: Bucket\n{indent}Virtual address: {vadr}\n{indent}Offset: {off}\n{indent}Key:\n{zend_string}{indent}Value:\n{zval}\n'.format(
                    **args)
            else:
                result = '{indent}Type: Bucket\n{indent}Virtual address: {vadr}\n{indent}Offset: {off}\n{indent}Index: {index_hash}({h_index_hash})\n{zval}\n'.format(
                    **args)
        return result

    def __repr__(self):
        return self.__get_data_to_print()

    def __str__(self):
        return self.__get_data_to_print()


if __name__ == "__main__":
    # filename = '../dumps/php7/core.7058'
    filename = '../dumps/php7/core.3175'
    from php7.zval_struct import ZvalStruct
    from php7.at_start.entry import init, get_glob_segments

    init(filename)

    # addr = 0x563dc6bdc4e0
    # print(Bucket(addr))

    # addr = 0x7f7c2865ca40 - 9 * 32  # Bucket
    # print(Bucket(addr))
    # addr = 0x7f7c2865ca40 - 8 * 32  # Bucket
    # print(Bucket(addr))
    # addr = 0x7f7c2865ca40 - 7 * 32  # Bucket
    # print(Bucket(addr))
    # addr = 0x7f7c2865ca40 - 6 * 32  # Bucket
    # print(Bucket(addr))
    # addr = 0x7f7c2865ca40 - 5 * 32  # Bucket
    # print(Bucket(addr))
    # addr = 0x7f7c2865ca40 - 4 * 32  # Bucket
    # print(Bucket(addr))
    # addr = 0x7f7c2865ca40 - 3 * 32  # Bucket
    # print(Bucket(addr))
    #
    # addr = 0x7f7c2865ca40 - 2 * 32  # Bucket
    # print(Bucket(addr))
    #
    # addr = 0x7f7c2865ca40 - 32  # Bucket
    # print(Bucket(addr))
    #
    # addr = 0x7f7c2865ca40  # Bucket
    # print(Bucket(addr))
    #
    # addr = 0x7f7c2865ca40 + 32  # Bucket
    # print(Bucket(addr))
    #
    # addr = 0x7f7c2865ca40 + 64  # Bucket
    # print(Bucket(addr))
    #
    # addr = 0x7f7c2865ca40 + 96  # Bucket
    # print(Bucket(addr))
    #
    # addr = 0x7f7c2865ca40 + 4*32  # Bucket
    # print(Bucket(addr))
    #
    # addr = 0x7f7c2865ca40 + 5 * 32  # Bucket
    # print(Bucket(addr))
    #
    # addr = 0x7f7c2865ca40 + 6 * 32  # Bucket
    # print(Bucket(addr))
    #
    # addr = 0x7f7c2865ca40 + 7 * 32  # Bucket
    # print(Bucket(addr))

    # addr = 0x7f7c28677380
    # print(Bucket(addr))
    #
    # addr = 0x7f7c2867b4e8
    # print(Bucket(addr))
    #
    # addr = 0x7f7c28688280
    # print(Bucket(addr))
    #
    # addr = 0x7f7c2868b080
    # print(Bucket(addr))

    #
    # addr =
    # print(Bucket(addr))
    #
    # addr =
    # print(Bucket(addr))
    #
    # addr =
    # print(Bucket(addr))
    #
    # addr =
    # print(Bucket(addr))
    #
    # addr =
    # print(Bucket(addr))
