from elf import get_offset, valid_address
from php7.at_start.entry import get_glob_segments, get_data
from php7.at_start.my_globals import *


class HashTableZvalue:
    def __init__(self, zvalue_raw_data, level=0) -> None:
        self.__zvalue_raw_data = zvalue_raw_data
        self.__level = level
        self.__buckets = []
        self.__address = unpack(zvalue_format_p, self.__zvalue_raw_data)[0]
        self.__valid_address = valid_address(self.__address, get_glob_segments())
        self.__valid = False
        if self.__valid_address:
            self.__offset = get_offset(self.__address, get_glob_segments())
            self.__data = get_data(self.__offset, array_size)
            self.__refcounted, (
                self.__flags, self.__nTableMask, self.__arData, self.__nNumUsed, self.__nNumOfElements,
                self.__nTableSize,
                self.__nInternalPointer, self.__nNextFreeElement), (
                self.__flags, self.__nApplyCount, self.__nIteratorsCount, self.__consistency) = self.__parse_hashtable()
            self.__valid = self.__check_valid()
            if self.is_valid():
                self.__buckets = self.__get_buckets()

    def __parse_hashtable(self) -> (tuple, tuple, tuple):
        refcounted = parse_refcounted(self.__data[:refcounted_size])
        return refcounted, unpack('IIQ4IQ', self.__data[refcounted_size:-8]), unpack('BBBB', self.__data[
                                                                                             refcounted_size:refcounted_size + 4])

    def __check_valid(self) -> bool:
        result = True
        result = result and self.__refcounted[1] == IS_ARRAY
        result = result and (valid_address(self.__arData, get_glob_segments()) or self.get_num_of_elem() == 0)
        result = result and (
        ((self.__nTableMask + self.__nTableSize) & 0xFFFFFFFF == 0) or (self.__nTableMask + 2) & 0xFFFFFFFF == 0)
        result = result and self.__nNumOfElements <= self.__nTableSize
        result = result and self.__nNumUsed < self.__nTableSize
        return result

    def is_valid(self):
        return self.__valid

    def get_num_of_elem(self):
        return self.__nNumOfElements

    def __get_buckets(self):
        result = []
        from php7.hashtables.bucket import Bucket
        for index in range(0, self.__nTableSize):
            zval = Bucket(self.__arData + index * bucket_size, self.__level + 1)
            if zval.is_valid():
                result.append(zval)
        return result

    def __get_data_to_print(self):
        indent = self.__level * spacer
        if not self.is_valid():
            return '{}Invalid Hashtable\n'.format(indent)
        arda = get_offset(self.__arData, get_glob_segments()) if valid_address(self.__arData,
                                                                               get_glob_segments()) else -1
        args = {'indent': indent, 'h_address': hex(self.__address), 'h_offset': hex(self.__offset),
                'address': int(self.__address), 'offset': int(self.__offset), "ardata": int(self.__arData),
                'hardata': hex(self.__arData), "oardata": int(arda),
                'ohardata': hex(arda),
                'carret': '{', 'ecarret': '}', 'noe': self.__nNumOfElements, 'lena': len(self.__buckets),
                'bucket_string': ''.join([str(x) for x in self.__buckets])}
        if values_only:
            # return '{indent}{carret}\n{bucket_string}{indent}{ecarret}'.format(**args)
            return '{indent}{carret}\n{bucket_string}{indent}{ecarret}{lena} of {noe}'.format(**args)
        else:
            return '{indent}Type: Hashtable\n{indent}Hashtable virtual address: {h_address}({address})\n{indent}Hashtable offset: {h_offset}({offset})\n{indent}Virtual address arData: {hardata}({ardata})\n{indent}Offset arData: {ohardata}({oardata})\n{bucket_string}'.format(
                **args)
            # return '{indent}Type: Hashtable\n{indent}Hashtable virtual address: {h_address}({address})\n{indent}Hashtable offset: {h_offset}({offset})\n{bucket_string}'.format(
            #     **args)

    def __repr__(self):
        return self.__get_data_to_print()

    def __str__(self):
        return self.__get_data_to_print()


if __name__ == "__main__":
    # filename = '../dumps/php7/core.7058'
    # filename = '../dumps/php7/core.3175'
    # filename = '../dumps/php7/core.6128'
    filename = '/home/martin/core.11846'
    filename = '../../dumps/php7/core.5834'
    filename = '../../dumps/php7/core.28039'
    filename = '../../dumps/php7/core.5769'

    from php7.at_start.entry import init
    from elf import get_all_addresses
    from struct import pack

    init(filename)

    # # hashtable offset 0x2b7494 na 7f7c286553f0
    # addr = get_address_from_offset(offset, get_glob_segments())
    # print(ZvalStruct(addr))

    # for address in get_all_addresses(get_glob_segments()):
    #     zval = ZvalStruct(address, 0)
    #     if zval.is_valid() and zval.get_type() == IS_ARRAY:
    #         print(zval)
    #     del zval
    for address in get_all_addresses(get_glob_segments()):
        zval = HashTableZvalue(pack('Q', address))
        if zval.is_valid():
            print(zval)
        del zval
