from struct import pack

from elf import valid_address, get_offset
from php7.at_start.entry import get_glob_segments, get_data, init
from php7.at_start.my_globals import *


class ZendClassEntry:
    def __init__(self, raw_data, level=0) -> None:
        self.__zvalue_raw_data = raw_data
        self.__level = level
        self.__address = unpack(zvalue_format_p, self.__zvalue_raw_data)[0]
        self.__valid_address = valid_address(self.__address, get_glob_segments())
        self.__valid = False
        self.__name, self.__default_properties_table, self.__default_static_members_table, self.__constants_table, self.__properties_info, self.__function_table, self.__static_members_table = None, None, None, None, None, None, None
        if self.__valid_address:
            self.__offset = get_offset(self.__address, get_glob_segments())
            self.__data = get_data(self.__offset, 64)
            self.__type, (
                self.__name_p, self.__parent, self.__refcount, self.__flags, self.__dpc, self.__dsmc, self.__dptp,
                self.__dsmtp, self.__smtp) = self.__parse_class_entry()
            from php7.string_zvalue import StringZvalue
            from php7.hashtables.hashtable_zvalue import HashTableZvalue
            from php7.zval_struct import ZvalStruct
            for i in [1]:
                if valid_address(self.__name_p, get_glob_segments()):
                    self.__name = StringZvalue(self.__data[8:16], self.__level + 1)
                    if not self.__name.is_valid():
                        continue
                if valid_address(self.__smtp, get_glob_segments()):
                    self.__static_members_table = ZvalStruct(self.__smtp, self.__level + 1)
                    if not self.__static_members_table.is_valid():
                        continue
                if self.__dpc and valid_address(self.__dptp, get_glob_segments()):
                    self.__default_properties_table = self.__create_zval_array(self.__dptp, self.__dpc)

                if valid_address(self.__address + 64, get_glob_segments()):
                    self.__function_table = HashTableZvalue(pack('Q', self.__address + 64), self.__level + 1)
                    if not self.__function_table.is_valid():
                        continue

                if valid_address(self.__address + 64 + 56, get_glob_segments()):
                    self.__properties_info = HashTableZvalue(pack('Q', self.__address + 64 + 56), self.__level + 1)
                    if not self.__properties_info.is_valid():
                        continue

                if self.__dsmc and valid_address(self.__dsmtp, get_glob_segments()):
                    self.__default_static_members_table = self.__create_zval_array(self.__dsmtp, self.__dsmc)

                if valid_address(self.__address + 64 + 2 * 56, get_glob_segments()):
                    self.__constants_table = HashTableZvalue(pack('Q', self.__address + 64 + 2 * 56), self.__level + 1)
                    if not self.__constants_table.is_valid():
                        continue
            self.__valid = self.__check_valid()

    def __parse_class_entry(self):
        return unpack('b', self.__data[:1])[0], unpack('QQiIiiQQQ', self.__data[8:])

    def __create_zval_array(self, address, count):
        result = []
        from php7.zval_struct import ZvalStruct
        for index in range(count):
            zval_address = address + index * zval_size
            if valid_address(zval_address, get_glob_segments()):
                zval = ZvalStruct(zval_address, self.__level + 1)
                if zval.is_valid():
                    result.append(zval)
        return result

    def is_valid(self):
        return self.__valid

    def __check_valid(self):
        result = True
        result = result and (self.__name is not None and self.__name.is_valid())
        result = result and (self.__parent == 0 or valid_address(self.__parent, get_glob_segments()))
        result = result and (self.__default_properties_table is not None or self.__dpc == 0)
        result = result and (self.__default_static_members_table is not None or self.__dsmc == 0)
        result = result and (
            self.__smtp == 0 or (self.__static_members_table is not None and self.__static_members_table.is_valid()))
        result = result and (self.__constants_table is not None and self.__constants_table.is_valid())
        result = result and (self.__properties_info is not None and self.__properties_info.is_valid())
        result = result and (self.__function_table is not None and self.__function_table.is_valid())
        return result

    def __get_data_to_print(self):
        indent = self.__level * spacer
        if not self.is_valid():
            return '{}Invalid Class entry\n'.format(indent)
        dpt = '' if self.__default_properties_table is None else '\n'.join(
            [str(x) for x in self.__default_properties_table])
        dsmt = '' if self.__default_static_members_table is None else '\n'.join(
            [str(x) for x in self.__default_static_members_table])
        args = {'indent': indent, 'h_address': hex(self.__address), 'h_offset': hex(self.__offset),
                'address': int(self.__address), 'offset': int(self.__offset), 'string': self.__name.get_string(),
                'name': self.__name, 'ft': self.__function_table, 'pi': self.__properties_info,
                'ct': self.__constants_table, 'dpt': dpt,
                'dsmt': dsmt, 'smt': self.__static_members_table}
        if values_only:
            return '{indent}Class entry:\n{indent}Name: {string}\n{indent}Default properties table:\n{dpt}\n{indent}Default static members table:\n{dsmt}{indent}Function table:\n{ft}\n{indent}Properties  info:\n{pi}\n{indent}Constants table:\n{ct}\n'.format(
                **args)
        else:
            return '{indent}Type: Class entry\n{indent}Class entry virtual address: {h_address}({address})\n{indent}Class entry offset: {h_offset}({offset})\n{indent}Default properties table:\n{dpt}{indent}Default static members table:\n{dsmt}{indent}Function table:\n{ft}{indent}Properties  info:\n{pi}{indent}Constants table:\n{ct}\n'.format(
                **args)

    def __str__(self) -> str:
        return self.__get_data_to_print()

    def __repr__(self) -> str:
        return self.__get_data_to_print()

    def get_properties_info_size(self):
        return self.__properties_info.get_num_of_elem()


if __name__ == "__main__":
    # filename = '../dumps/php7/core.7058'
    # filename = '../dumps/php7/core.3175'
    filename = '../dumps/php7/core.6128'
    init(filename)

    # for address in get_all_addresses(get_glob_segments()):
    # for address in [0x7f013cc730b0]:
    for address in [0x7f013cc03018]:
        zval = ZendClassEntry(pack('Q', address))
        if zval.is_valid():
            print(zval, hex(address))
        del zval
