from struct import pack

from elf import valid_address, get_offset
from php7.at_start.entry import get_glob_segments, get_data, init
from php7.at_start.my_globals import *


class ObjectZvalue:
    def __init__(self, raw_data, level=0) -> None:
        self.__zvalue_raw_data = raw_data
        self.__level = level
        self.__address = unpack(zvalue_format_p, self.__zvalue_raw_data)[0]
        self.__valid_address = valid_address(self.__address, get_glob_segments())
        self.__valid = False
        if self.__valid_address:
            self.__offset = get_offset(self.__address, get_glob_segments())
            self.__data = get_data(self.__offset, 40)
            self.__refcounted, (
                self.__class_entry_address, self.__obj_handlers, self.__properties_address) = self.__parse_object()
            from php7.objects.class_entry_zvalue import ZendClassEntry
            from php7.hashtables.hashtable_zvalue import HashTableZvalue
            if valid_address(self.__properties_address, get_glob_segments()):
                self.__properties = HashTableZvalue(pack('Q', self.__properties_address), self.__level + 1)
                if valid_address(self.__class_entry_address, get_glob_segments()):
                    self.__class_entry = ZendClassEntry(pack('Q', self.__class_entry_address), self.__level + 1)
                    if self.__properties.is_valid():
                        self.__properties_table = self.get_properties_table()
                        self.__valid = self.__check_valid()

    def __parse_object(self):
        return parse_refcounted(self.__data[:8]), unpack('QQQ', self.__data[16:])

    def is_valid(self):
        return self.__valid

    def __check_valid(self):
        return self.__refcounted[
                   1] == IS_OBJECT and self.__class_entry.is_valid() and self.__properties.is_valid() and valid_address(
            self.__obj_handlers, get_glob_segments())
        # return self.__refcounted[1] == IS_OBJECT and valid_address(self.__class_entry_address, get_glob_segments()) and self.__properties.is_valid() and valid_address(self.__obj_handlers, get_glob_segments())

    def __get_data_to_print(self):
        indent = self.__level * spacer
        indent2 = (self.__level + 1) * spacer

        if not self.is_valid():
            return '{}Invalid Object\n'.format(indent)
        args = {'indent': indent, 'indent2': indent2, 'h_address': hex(self.__address), 'h_offset': hex(self.__offset),
                'address': int(self.__address), 'offset': int(self.__offset), 'prop': self.__properties,
                'ent': self.__class_entry, 'pt': '\n'.join([str(x) for x in self.__properties_table])}
        if values_only:
            return '{indent}Object:\n{indent2}Properties:\n{prop}\n{indent2}Properties table:\n{pt}\n{ent}'.format(
                **args)
        else:
            return '{indent}Type: Object\n{indent}Object virtual address: {h_address}({address})\n{indent}Object offset: {h_offset}({offset})\n{indent}Properties:\n{prop}\n{ent}\n{indent}Properties table:\n{pt}\n'.format(
                **args)

    def __str__(self) -> str:
        return self.__get_data_to_print()

    def __repr__(self) -> str:
        return self.__get_data_to_print()

    def get_properties_table(self):
        from php7.zval_struct import ZvalStruct
        address = self.__address + 40
        count = self.__class_entry.get_properties_info_size()
        result = []
        for index in range(count):
            zval_address = address + index * zval_size
            if valid_address(zval_address, get_glob_segments()):
                zval = ZvalStruct(zval_address, self.__level + 2)
                if zval.is_valid():
                    result.append(zval)
        return result


if __name__ == "__main__":
    # filename = '../dumps/php7/core.7058'
    filename = '../dumps/php7/core.6128'
    # filename = '../dumps/php7/core.3175'
    from php7.zval_struct import ZvalStruct

    init(filename)

    # for address in get_all_addresses(get_glob_segments()):
    for address in [0x7f013cc13090, 0x7f013cc13170]:
        zval = ZvalStruct(address, 0)
        if zval.is_valid() and zval.get_type() == IS_OBJECT:
            print(zval)
        del zval
