from elf import valid_address
from php7.at_start.entry import get_glob_segments
from php7.at_start.my_globals import *
from php7.constant_in_table import ConstantInTable
from php7.functions.function_zvalue import FunctionZvalue
from php7.objects.class_entry_zvalue import ZendClassEntry


class PointerZvalue:
    def __init__(self, raw_data, level=0) -> None:
        self.__zvalue_raw_data = raw_data
        self.__level = level
        self.__address = unpack(zvalue_format_p, self.__zvalue_raw_data)[0]
        self.__valid_address = valid_address(self.__address, get_glob_segments())
        self.__valid = False
        self.__values = []
        if self.__valid_address:
            zend_function = FunctionZvalue(self.__zvalue_raw_data, self.__level)
            self.__values.append(zend_function)

            class_entry = ZendClassEntry(self.__zvalue_raw_data, self.__level)
            if class_entry is not None and class_entry.is_valid():
                self.__values.append(class_entry)
            constant = ConstantInTable(self.__zvalue_raw_data, self.__level)
            # if constant.is_valid():
            self.__values.append(constant)
            self.__valid = self.check_valid()

    def check_valid(self):
        for structure in self.__values:
            if structure.is_valid():
                return True
        return False

    def is_valid(self):
        return len(self.__values) > 0

    def get_data_to_print(self):
        indent = self.__level * spacer
        if not self.__valid:
            return '{}Invalid pointer zend value\n'.format(indent)
        return '\n'.join([str(x) for x in self.__values if x.is_valid()])

    def __repr__(self) -> str:
        return self.get_data_to_print()

    def __str__(self) -> str:
        return self.get_data_to_print()
