from elf import valid_address, get_offset, get_all_addresses
from php7.at_start.entry import get_glob_segments, get_data, init
from php7.at_start.my_globals import *


class ReferenceZvalue:
    def __init__(self, raw_data, level=0) -> None:
        self.__zvalue_raw_data = raw_data
        self.__level = level
        self.__address = unpack(zvalue_format_p, self.__zvalue_raw_data)[0]
        self.__valid_address = valid_address(self.__address, get_glob_segments())
        self.__valid = False
        if self.__valid_address:
            self.__zval_address = self.__address + refcounted_size
            self.__offset = get_offset(self.__address, get_glob_segments())
            self.__data = get_data(self.__offset, reference_size)
            self.__refcounted, self.__zval = self.__parse_reference()
            self.__valid = self.__check_valid()

    def __parse_reference(self):
        from php7.zval_struct import ZvalStruct
        return parse_refcounted(self.__data[:refcounted_size]), ZvalStruct(self.__zval_address,
                                                                           self.__level + 1)

    def is_valid(self):
        return self.__valid

    def __check_valid(self):
        return self.__refcounted[1] == IS_REFERENCE and self.__zval.is_valid()

    def __get_data_to_print(self):
        indent = self.__level * spacer
        if not self.is_valid():
            return '{}Invalid Reference\n'.format(indent)
        args = {'indent': indent, 'h_address': hex(self.__address), 'h_offset': hex(self.__offset),
                'address': int(self.__address), 'offset': int(self.__offset), 'zval': self.__zval}
        if values_only:
            return '{indent}Referenced:\n{zval}'.format(**args)
        else:
            return '{indent}Type: Reference\n{indent}Reference virtual address: {h_address}({address})\n{indent}Reference offset: {h_offset}({offset})\n{zval}'.format(
                **args)

    def __str__(self) -> str:
        return self.__get_data_to_print()

    def __repr__(self) -> str:
        return self.__get_data_to_print()


if __name__ == "__main__":
    # filename = '../dumps/php7/core.7058'
    filename = '../dumps/php7/core.6128'
    # filename = '../dumps/php7/core.3175'
    from php7.zval_struct import ZvalStruct

    init(filename)

    for address in get_all_addresses(get_glob_segments()):
        zval = ZvalStruct(address, 0)
        if zval.is_valid() and zval.get_type() == IS_REFERENCE:
            print(zval)
        del zval
