from elf import valid_address, get_offset, get_all_addresses
from php7.at_start.entry import get_glob_segments, get_data
from php7.at_start.my_globals import *


class ResourceZvalue:
    def __init__(self, raw_data, level=0) -> None:
        self.__zvalue_raw_data = raw_data
        self.__level = level
        self.__address = unpack(zvalue_format_p, self.__zvalue_raw_data)[0]
        self.__valid_address = valid_address(self.__address, get_glob_segments())
        self.__valid = False
        if self.__valid_address:
            self.__offset = get_offset(self.__address, get_glob_segments())
            self.__data = get_data(self.__offset, resource_size)
            self.__refcounted, (
                self.__handle, self.__resource_type, self.__resourse_data_pointer) = self.__parse_resource()
            self.__valid = self.__check_valid()

    def __parse_resource(self):
        return parse_refcounted(self.__data[:refcounted_size]), unpack('iiQ', self.__data[refcounted_size:])

    def is_valid(self):
        return self.__valid

    def __check_valid(self):
        return self.__refcounted[1] == IS_RESOURCE and valid_address(self.__resourse_data_pointer, get_glob_segments())

    def __get_data_to_print(self):
        indent = self.__level * spacer
        if not self.is_valid():
            return '{}Invalid Resource\n'.format(indent)
        res_offset = get_offset(self.__resourse_data_pointer, get_glob_segments())
        args = {'indent': indent, 'h_address': hex(self.__address), 'h_offset': hex(self.__offset),
                'address': int(self.__address), 'offset': int(self.__offset), 'hrva': hex(self.__resourse_data_pointer),
                'rva': int(self.__resourse_data_pointer), 'hrdo': hex(res_offset), 'rdo': int(res_offset),
                'type': self.__resource_type}
        if values_only:
            return '{indent}Resource:\n{indent}{type} {hrva}'.format(**args)
        else:
            return '{indent}Type: Resource\n{indent}Resource type: {type}\n{indent}Resource data virtual address: {hrva}({rva})\n{indent}Resource data offset: {hrdo}({rdo})\n{indent}Zend Resource virtual address: {h_address}({address})\n{indent}Zend Resource offset: {h_offset}({offset})\n'.format(
                **args)

    def __str__(self) -> str:
        return self.__get_data_to_print()

    def __repr__(self) -> str:
        return self.__get_data_to_print()


if __name__ == "__main__":
    # filename = '../dumps/php7/core.7058'
    # filename = '../dumps/php7/core.3175'
    filename = '../dumps/php7/core.6128'
    from php7.zval_struct import ZvalStruct
    from php7.at_start.entry import init

    init(filename)

    for address in get_all_addresses(get_glob_segments()):
        zval = ZvalStruct(address, 0)
        if zval.is_valid() and zval.get_type() == IS_RESOURCE:
            print(zval)
        del zval
