from struct import pack

from elf import valid_address, get_offset, get_all_addresses
from php7.at_start.entry import init, get_glob_segments, get_data
from php7.at_start.my_globals import *


class StringZvalue:
    def __init__(self, zvalue_raw_data, level=0, address=-1) -> None:
        # def __init__(self, zval_struct, level=0, address = -1) -> None:
        # def __init__(self, zval_struct: ZvalStruct, level=0) -> None:
        self.__valid = False
        self.__level = level
        self.__zvalue_raw_data = zvalue_raw_data
        self.__zend_string_data, self.__string = '', ''
        if address == -1:
            self.__address = unpack(zvalue_format_p, zvalue_raw_data)[0]
        self.__valid_address = valid_address(self.__address, get_glob_segments())
        if self.__valid_address:
            self.__offset = get_offset(self.__address, get_glob_segments())
            self.__zend_string_data = get_data(self.__offset, 24)
            self.__string_refcounted, self.__string_hash, self.__length = self.__parse_string()
            # if self.__string_refcounted[1] == IS_STRING :
            # if self.__string_refcounted[1] == IS_STRING and self.__length > 0:
            # if self.__string_refcounted[1] == IS_STRING and self.__length > 0 and self.__length < deprecation_size:
            if self.__string_refcounted[1] == IS_STRING and self.__length < deprecation_size:
                # if self.__string_refcounted[1] == IS_STRING and self.__string_refcounted[0] > 0:
                # if self.__string_refcounted[1] == IS_STRING and self.__string_refcounted[0] > 0 and self.__length > 0:
                self.__valid = True
                self.load_string()

    def __parse_string(self):
        string_hash, length = unpack('QQ', self.__zend_string_data[8:])
        return parse_refcounted(self.__zend_string_data[:8]), string_hash, length

    def is_valid(self):
        return self.__valid

    def load_string(self):
        from php7.at_start.entry import get_data
        self.__string = get_data(self.__offset + 24, self.__length)

    def get_string(self):
        return self.__string

    def get_length(self):
        return self.__length

    def get_data_to_print(self):
        indent = self.__level * spacer
        if not self.is_valid():
            return '{}Invalid String\n'.format(indent)
        if values_only:
            return '{indent}{string}'.format(string=self.__string, indent=indent)
        args = {'indent': indent, 'string': self.__string, 'length': self.__length, 'hash': hex(self.__string_hash),
                'refcount': self.__string_refcounted[0], 'zsva': int(self.__address), 'hzsva': hex(self.__address),
                'sva': self.__address + 24, 'hsva': hex(self.__address + 24), 'zsvo': int(self.__offset),
                'hzsvo': hex(self.__offset), 'svo': self.__offset + 24, 'hsvo': hex(self.__offset + 24)
                }
        result = '{indent}Type: String\n{indent}Value: {string}\n{indent}Length: {length}\n{indent}Hash: {hash}\n{indent}Zend string virtual address: {hzsva}({zsva})\n{indent}Zend string virtual offset: {hzsvo}({hzsvo})\n{indent}String virtual address: {hsva}({sva})\n{indent}String virtual offset: {hsvo}({svo})\n{indent}Refcount: {refcount}\n'.format(
            **args)
        return result

    def __str__(self) -> str:
        return self.get_data_to_print()

    def __repr__(self) -> str:
        return self.get_data_to_print()


if __name__ == "__main__":
    # filename = '../dumps/php7/core.7058'
    # filename = '../dumps/php7/core.3175'
    filename = '../dumps/php7/core.6128'
    filename = '../core.5834'
    filename = '../dumps/php7/core.5769'

    init(filename)

    from php7.zval_struct import ZvalStruct

    # 0x2c9164   0x7f7c286670c0 ref at 0x274144 0x7f7c286120a0
    # addr = 0x7f7c286670c0
    # addr = 0x7f7c286120a0
    # print(ZvalStruct(addr))

    for address in get_all_addresses(get_glob_segments()):
        zval = StringZvalue(pack('Q', address))
        if zval.is_valid():
            print(zval)
        del zval

        # for address in get_all_addresses(get_glob_segments()):
        #     zval = ZvalStruct(address, 0)
        #     if zval.is_valid() and zval.get_type() == IS_STRING:
        #         print(zval)
        #     del zval
