from struct import unpack, pack

from elf import valid_address, get_offset
from php7.at_start.entry import get_glob_segments, get_data
from php7.at_start.my_globals import zvalue_format_p, execute_data_partial, spacer, values_only
from php7.functions.function_zvalue import FunctionZvalue
from php7.functions.zend_op_array import ZendOpArray
from php7.hashtables.hashtable_zvalue import HashTableZvalue


class ZendExecuteData:
    def __init__(self, raw_data, level=0) -> None:
        self.__raw_data = raw_data
        self.__level = level
        self.__address = unpack(zvalue_format_p, self.__raw_data)[0]
        self.__valid = False
        self.__valid_address = valid_address(self.__address, get_glob_segments())
        self.__function = None
        self.__prev_execute_data = None
        self.__symbol_table = None
        if self.__valid_address:
            self.__offset = get_offset(self.__address, get_glob_segments())
            self.__data = get_data(self.__offset, execute_data_partial)
            (self.__opline, self.__call, self.__return_value, self.__func_p), (
                self.__prev_execute_data_p, self.__symbol_table_p) = self.__parse_execute_data()
            if self.__partial_check():
                self.__function = ZendOpArray(pack('Q', self.__func_p), self.__level + 1)
                self.__symbol_table = HashTableZvalue(pack('Q', self.__symbol_table_p), self.__level + 1)
                if self.__prev_execute_data_p is not 0:
                    self.__prev_execute_data = ZendExecuteData(pack('Q', self.__prev_execute_data_p), self.__level)
                self.__valid = self.__check_valid()

    def is_valid(self):
        return self.__valid

    def __parse_execute_data(self):
        return unpack('4Q', self.__data[:32]), unpack('2Q', self.__data[-16:])

    def __partial_check(self):
        result = True
        result = result and (self.__opline == 0 or valid_address(self.__opline, get_glob_segments()))
        result = result and (self.__call == 0 or valid_address(self.__call, get_glob_segments()))
        result = result and (self.__return_value == 0 or valid_address(self.__return_value, get_glob_segments()))
        return result

    def __check_valid(self):
        result = True
        result = result and (self.__prev_execute_data_p == 0 or self.__prev_execute_data.is_valid())
        # result = result and (self.__symbol_table_p == 0 or self.__symbol_table.is_valid())
        result = result and (self.__function.is_valid())
        return result

    def __get_data_to_print(self):
        indent1 = self.__level * spacer
        indent2 = (self.__level + 1) * spacer
        if not self.is_valid():
            return '{}Invalid execute data\n'.format(indent1)
        # prev = hex(self.__prev_execute_data_p) if self.__prev_execute_data_p == 0 else self.__prev_execute_data
        prev = self.__prev_execute_data
        args = {'indent1': indent1, 'indent2': indent2, 'function': self.__function, 'symbol': self.__symbol_table,
                'prev': prev, 'ushp': hex(self.__symbol_table_p)}
        if values_only:
            return '{indent1}Execute data{ushp}\n{function}{symbol}\nPrev: {prev}'.format(**args)
        else:
            return '{indent1}Execute data\n{function}{symbol}{prev}'.format(**args)

    def __str__(self) -> str:
        return self.__get_data_to_print()

    def __repr__(self) -> str:
        return self.__get_data_to_print()


if __name__ == "__main__":
    filename = '../dumps/php7/core.6128'
    filename = '../core.28039'
    from php7.at_start.entry import init
    from elf import get_all_addresses

    init(filename)

    for address in get_all_addresses(get_glob_segments()):
        zval = ZendExecuteData(pack('Q', address))
        if zval.is_valid():
            print(zval, hex(address), '\n==========================================')
        del zval
