from elf import get_offset, get_all_addresses, get_address_from_offset
from php7.at_start.entry import get_data
from php7.at_start.my_globals import *
from php7.double_zvalue import DoubleZvalue
from php7.false_zvalue import FalseZvalue
from php7.hashtables.hashtable_zvalue import HashTableZvalue
from php7.indirect_zvalue import IndirectZvalue
from php7.invalid_zvalue import InvalidZvalue
from php7.long_zvalue import LongZvalue
from php7.null_zvalue import NullZvalue
from php7.objects.object_zvalue import ObjectZvalue
from php7.pointer_zvalue import PointerZvalue
from php7.reference_zvalue import ReferenceZvalue
from php7.resource_zvalue import ResourceZvalue
from php7.string_zvalue import StringZvalue
from php7.true_zvalue import TrueZvalue
from php7.undef_zvalue import UndefinedZvalue
from php7.at_start.entry import get_glob_segments


class ZvalStruct:
    def __init__(self, address, level=0, offset=-1):
        self.__level = level
        self.__valid = False
        self.__zvalue = None
        if offset == -1:
            self.__address = address
            self.__offset = get_offset(address, get_glob_segments())
        else:
            self.__offset = offset
            self.__address = get_address_from_offset(offset, get_glob_segments())
        self.__data = get_data(self.get_zval_struct_offset(), zval_size)
        self.__zvalue_raw_data, self.__u1, self.__u2, self.__type, self.__type_flags, self.__const_flags, self.__reserved = self.parse_zval_struct()
        self.__valid_type = self.__check_type()
        if self.__valid_type:
            self.__zvalue = self.__create_zvalue()
            if self.__zvalue is not None and self.__zvalue.is_valid():
                self.__valid = True

    def get_zval_struct_address(self):
        return self.__address

    def get_zval_struct_offset(self):
        return self.__offset

    def get_level(self):
        return self.__level

    def is_valid(self):
        return self.__valid

    def get_zvalue_raw_data(self):
        return self.__zvalue_raw_data

    def get_type(self):
        return self.__type

    def __create_zvalue(self):
        result = None
        raw_data = self.get_zvalue_raw_data()
        if self.__type == IS_UNDEF:
            result = UndefinedZvalue(raw_data, self.__level)
        elif self.__type == IS_NULL:
            result = NullZvalue(raw_data, self.__level)
        elif self.__type == IS_FALSE:
            result = FalseZvalue(raw_data, self.__level)
        elif self.__type == IS_TRUE:
            result = TrueZvalue(raw_data, self.__level)
        elif self.__type == IS_LONG:
            result = LongZvalue(raw_data, self.__level)
        elif self.__type == IS_DOUBLE:
            result = DoubleZvalue(raw_data, self.__level)
        elif self.__type == IS_STRING:
            result = StringZvalue(raw_data, self.__level)
        elif self.__type == IS_ARRAY:
            result = HashTableZvalue(raw_data, self.__level)
        elif self.__type == IS_OBJECT:
            pass
            result = ObjectZvalue(raw_data, self.__level)
        elif self.__type == IS_RESOURCE:
            result = ResourceZvalue(raw_data, self.__level)
        elif self.__type == IS_REFERENCE:
            result = ReferenceZvalue(raw_data, self.__level)
        elif self.__type == IS_CONSTANT:
            pass
            result = StringZvalue(raw_data, self.__level)
            # result = ConstantZvalue(raw_data, self.__level)
        elif self.__type == IS_CONSTANT_AST:
            pass
        elif self.__type == IS_INDIRECT:
            result = IndirectZvalue(raw_data, self.__level)
        elif self.__type == IS_PTR:
            result = PointerZvalue(raw_data, self.__level)
            pass
        else:
            result = InvalidZvalue()
        return result

    def __repr__(self):
        indent = self.__level * spacer
        if not self.__valid:
            return indent + 'Invalid zval\n'
        if values_only:
            return self.__zvalue.__repr__()
        return self.__zvalue.__repr__() + self.get_info()

    def __str__(self):
        indent = self.__level * spacer
        if not self.__valid:
            return indent + 'Invalid zval\n'
        if values_only:
            return self.__zvalue.__str__()
        return self.__zvalue.__str__() + self.get_info()

    def get_info(self):
        indent = self.__level * spacer
        args = {'indent': indent, 'zsva': self.__address, 'zso': self.__offset, 'hzsva': hex(self.__address),
                'hzso': hex(self.__offset)}
        return '{indent}Zval virtual address: {hzsva}({zsva})\n{indent}Zval offset: {hzso}({zso})\n'.format(
            **args)

    def parse_zval_struct(self):
        v = self.__data[:8]
        u1, u2 = unpack('II', self.__data[8:])
        t, tf, cf, r = unpack('BBBB', self.__data[8:12])
        return v, u1, u2, t, tf, cf, r

    def __check_type(self):
        return self.__type in [IS_UNDEF, IS_FALSE, IS_TRUE, IS_NULL, IS_LONG, IS_DOUBLE, IS_STRING, IS_OBJECT, IS_ARRAY,
                               IS_RESOURCE, IS_REFERENCE, IS_CONSTANT, IS_INDIRECT, IS_PTR]


if __name__ == '__main__':
    filename = '../dumps/php7/core.6128'
    filename = '/home/martin/sedmicka/core.5769'
    filename = '../dumps/php7/core.9169'

    # f = open(filename, 'rb')
    num = 0
    from php7.at_start.entry import get_glob_segments, init

    init(filename)

    for address in get_all_addresses(get_glob_segments()):
        zval = ZvalStruct(address)
        # if zval.is_valid() and zval.get_type() in [IS_CONSTANT]:
        if zval.is_valid() and zval.get_type() in [IS_STRING, IS_OBJECT, IS_ARRAY, IS_RESOURCE, IS_REFERENCE,
                                                   IS_INDIRECT]:
            print(zval)
            num += 1

    print(num)
    pass
